// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//  Giuseppe Vacanti (cosine science & computing bv)
//  March 18, 2003
// 
//   $Id: Observation_namespace.cc,v 1.2 2003/08/26 14:46:43 gvacanti Exp $
#include<Observation.h>

#include<ExceptionErrorHandler.h>
#include<SmartDal>
#include<HkNamer.h>
#include<Hkparameter.h>
#include<Xmm.h>

#include<algorithm>
#include<ostream>
#include<algorithm>
#include<iterator>
#include<exception>

#include<boost/lexical_cast.hpp>

namespace {
  using namespace std;

  class SortByHkFileType {
  public:
    bool operator()(const Hkparameter * rh, const Hkparameter * lh){
      return rh->hktype() < lh->hktype();
    }
  };
  
  ostream & operator<<(ostream & os, pair<string,string> & p) {
    os << p.first << ":" << p.second;
    return os;
  }
  

  // This is a helper class to write translators of parameter
  // values. The problem I'm solving here is the following. Some of
  // the values extracted from the HK are not suitable to be directly
  // used in the SAS. The reasons are:

  // - HK value does not mean anything for end user. 
  //   Example: the MOS gain is CHAINNORNOD0 to mean NORMAL
  // - HK value is mispelled.
  //   Example: the PN filter position can be NO STOP POS or NO STOP POS.
  //   (with the additional .).
  // - HK value does not match the names used in the SAS.
  //   Example: PN filter FILTC-MEDIUM is just called MEDIUM in the SAS,
  //   the same name for PN and MOS.
  // - HK value numeric but meaningful value should be string.
  //   Example: MOS filter positions are integers, but in the SAS 
  //   they should be strings.
  // - HK value floating point in a range that should map to a single string.

  // Most translations go from one set of values to another set of
  // values, one to one. These are handled with maps. Some are more
  // complex (floating point in range), and they have to be handled
  // with specialized code. They are described below.
  
  template<typename T> class Translator {
  public:

    class TranslatorException {
    public:
      TranslatorException() {}
    };

    
    Translator() {}
    Translator(const map<string,string> & names,
	       const map<T,string> & dictionary)
      : names_(names), dictionary_(dictionary) {}
    
    void translate(OdfIngest::Exposure & exp) {
      using namespace OdfIngest;
      typedef map<string,string>::const_iterator MI;
      for(MI K = names_.begin(); K != names_.end(); ++K){
	const string & name = (*K).first;
	if(exp.has_ippv(name)){
	  Ippv p = exp.get_ippv(name);
	  string new_value;
	  bool invalid = true;
	  try {
	    // I cannot use the templated value mething in Ippv
	    // because that one cannot cope with T=string.
	    T value = boost::lexical_cast<T>(p.value());
	    new_value = get_new_value(value);
	    invalid = false;
	  }
	  catch(boost::bad_lexical_cast &) {
	    errstr << "I could not perform a lexical conversion on "
		   << p.name() << " with value " << p.value()
		   << ". The resulting housekeeping parameter " << names_[name]
		   << " will not be created."
		   << Sas::warning("HousekeepingExtractionFailed");
	  }
	  catch(TranslatorException & e) {
	    errstr << "I could not perform the translation of the housekeeping parameter "
		   << p.name() << " with value " << p.value()
		   << ". The resulting housekeeping parameter "
		   << names_[name] << " will not be created."
		   << Sas::warning("HousekeepingExtractionFailed");
	  }
	  exp.insert_ippv(Ippv(names_[name], new_value, invalid));
	} else
	  errstr << "The exposure " << exp.identifier()
		 << " does not contain any parameter named " << name
		 << ". The corresponding housekeeping parameter "
		 << names_[name] << " will not be created."
		 << Sas::warning("HousekeepingExtractionFailed");
      }
    }

    string get_new_value(const T & old_value) {
      if(dictionary_.find(old_value) == dictionary_.end())
	throw TranslatorException();
      return dictionary_[old_value];
    }
    
  private:
    map<string,string> names_;
    map<T,string> dictionary_;
  };


  // Here are the concrete implementations of the translators. They
  // make use of the helper class above.

  class ActualTranslator {
  public:
    ActualTranslator() {}
    virtual void operator()(OdfIngest::Exposure &) {
      errstr << "ActualTranslator::operator() invoked. This is a bug."
	     << Sas::error("WrongOperator");
    }
    virtual ~ActualTranslator() {}
  };

  class EmosGainTranslator : public ActualTranslator {
  public:
    EmosGainTranslator() {
      map<string,string> dictionary;
      dictionary["CHAINNORNOD0"] = "NORMAL";
      dictionary["CHAINNORNOD1"] = "NORMAL";
      dictionary["CHAIN/10NOD0"] = "LOW";
      dictionary["CHAIN/10NOD1"] = "LOW";
      map<string,string> names;
      names["GAIN_CCD_1_RAW"] = "GAIN_CCD_1";
      names["GAIN_CCD_2_RAW"] = "GAIN_CCD_2";
      names["GAIN_CCD_3_RAW"] = "GAIN_CCD_3";
      names["GAIN_CCD_4_RAW"] = "GAIN_CCD_4";
      names["GAIN_CCD_5_RAW"] = "GAIN_CCD_5";
      names["GAIN_CCD_6_RAW"] = "GAIN_CCD_6";
      names["GAIN_CCD_7_RAW"] = "GAIN_CCD_7";
      translator_ = Translator<string>(names, dictionary);
    }
    void operator()(OdfIngest::Exposure & exp){
      translator_.translate(exp);
    }
  private:
    Translator<string> translator_;
  };

  class Emos1FilterTranslator : public ActualTranslator {
  public:
    Emos1FilterTranslator() {
      map<int,string> dictionary;
      dictionary[0] = Xmm::toString(Xmm::CLOSED);
      dictionary[1334] = Xmm::toString(Xmm::OPEN);
      dictionary[267] = Xmm::toString(Xmm::THIN1);
      dictionary[534] = Xmm::toString(Xmm::THIN2);
      dictionary[801] = Xmm::toString(Xmm::MEDIUM);
      dictionary[1067] = Xmm::toString(Xmm::THICK);
      dictionary[1580] = Xmm::toString(Xmm::CAL_CLOSED);

      dictionary[1314] = Xmm::toString(Xmm::CAL_OPEN);
      dictionary[247] = Xmm::toString(Xmm::CAL_THIN1);
      dictionary[514] = Xmm::toString(Xmm::CAL_THIN2);
      dictionary[781] = Xmm::toString(Xmm::CAL_MEDIUM);
      dictionary[1047] = Xmm::toString(Xmm::CAL_THICK);      
      map<string,string> names;
      names["FILTER_WHEEL_POSITION"] = "FILTER";
      translator_ = Translator<int>(names, dictionary);
    }
    void operator()(OdfIngest::Exposure & exp){
      translator_.translate(exp);
    }
  private:
    Translator<int> translator_;
  };

  class Emos2FilterTranslator : public ActualTranslator {
  public:
    Emos2FilterTranslator() {
      map<int,string> dictionary;
      dictionary[0] = Xmm::toString(Xmm::CLOSED);
      dictionary[1332] = Xmm::toString(Xmm::OPEN);
      dictionary[266] = Xmm::toString(Xmm::THIN1);
      dictionary[533] = Xmm::toString(Xmm::THIN2);
      dictionary[799] = Xmm::toString(Xmm::MEDIUM);
      dictionary[1066] = Xmm::toString(Xmm::THICK);
      dictionary[1580] = Xmm::toString(Xmm::CAL_CLOSED);
      dictionary[1312] = Xmm::toString(Xmm::CAL_OPEN);
      dictionary[246] = Xmm::toString(Xmm::CAL_THIN1);
      dictionary[513] = Xmm::toString(Xmm::CAL_THIN2);
      dictionary[779] = Xmm::toString(Xmm::CAL_MEDIUM);
      dictionary[1046] = Xmm::toString(Xmm::CAL_THICK);
      map<string,string> names;
      names["FILTER_WHEEL_POSITION"] = "FILTER";
      translator_ = Translator<int>(names, dictionary);
    }
    void operator()(OdfIngest::Exposure & exp){
      translator_.translate(exp);
    }
  private:
    Translator<int> translator_;
  };

  class OmFilterTranslator : public ActualTranslator {
  public:
    OmFilterTranslator() {
      map<int,string> dictionary;
      dictionary[1200] = Xmm::toString(Xmm::BLOCKED);
      dictionary[1400] = Xmm::toString(Xmm::V);
      dictionary[1600] = Xmm::toString(Xmm::MAGNI);
      dictionary[1800] = Xmm::toString(Xmm::U);
      dictionary[2000] = Xmm::toString(Xmm::B);
      dictionary[0] = Xmm::toString(Xmm::WHITE);
      dictionary[200] = Xmm::toString(Xmm::GRISM2);
      dictionary[400] = Xmm::toString(Xmm::UVW1);
      dictionary[600] = Xmm::toString(Xmm::UVM2);
      dictionary[800] = Xmm::toString(Xmm::UVW2);
      dictionary[1000] = Xmm::toString(Xmm::GRISM1);
      dictionary[2100] = Xmm::toString(Xmm::BARRED_U);
      dictionary[2200] = Xmm::toString(Xmm::FILTER_NONE);
      map<string,string> names;
      names["FILTER_WHEEL_POSITION"] = "FILTER";
      translator_ = Translator<int>(names, dictionary);
    }
    void operator()(OdfIngest::Exposure & exp){
      translator_.translate(exp);
    }
  private:
    Translator<int> translator_;
  };

  class EpnFilterWheelTranslator : public ActualTranslator {
  public:
    EpnFilterWheelTranslator() {
      map<string,string> dictionary;
      dictionary["OPEN"] = Xmm::toString(Xmm::OPEN);
      dictionary["FILTRD-THICK"] = Xmm::toString(Xmm::THICK);
      dictionary["FILTC-MEDIUM"] = Xmm::toString(Xmm::MEDIUM);
      // How many typos are there?? Is it THIN or THIN2??
      dictionary["FILTRB-THIN"] = Xmm::toString(Xmm::THIN2);
      dictionary["FILTRB-THIN2"] = Xmm::toString(Xmm::THIN2);
      dictionary["FILTRA-THIN1"] = Xmm::toString(Xmm::THIN1);
      dictionary["CLOSE"] = Xmm::toString(Xmm::CLOSED);
      dictionary["NO STOP POS"] = Xmm::toString(Xmm::FILTER_NONE);
      dictionary["NO STOP POS."] = Xmm::toString(Xmm::FILTER_NONE);      
      map<string,string> names;
      names["FILTER_WHEEL_NAME"] = "FILTER_WHEEL";
      translator_ = Translator<string>(names, dictionary);
    }
    void operator()(OdfIngest::Exposure & exp){
      translator_.translate(exp);
    }
  private:
    Translator<string>  translator_;
  };



  // This translates the filter wheel position as given by the
  // potentiometer readout into a SAS compliant string. The algorithm
  // is takes from the XSCS software.
  class EpnFilterTranslator : public ActualTranslator {
  public:
    EpnFilterTranslator() {}
    
    void operator()(OdfIngest::Exposure & exp) {
      class InRange {
      public:
	InRange(float x) : x_(x) {}
	bool operator()(float lo, float up) const {
	  return (x_ > lo) && (x_ <= up);
	}
      private:
	float x_;
      };

      using namespace OdfIngest;
      string oname("FILTER_WHEEL_POSITION_POTENTIOMETER");
      string nname("FILTER_WHEEL_POSITION");
      if(exp.has_ippv(oname)) {
	OdfIngest::Ippv p = exp.get_ippv(oname);
	string new_value;
	bool invalid = true;
	try {
	  float v = p.value<float>(); // this can trowh a string
	  InRange ranger(v);
	  // These values taken from XSCS code F_FW_POS
	  if(ranger(42., 52.))
	    new_value = Xmm::toString(Xmm::CAL_CLOSED);
	  else if(ranger(52., 58.))
	    new_value = Xmm::toString(Xmm::CLOSED);
	  else if(ranger(102., 113.))
	    new_value = Xmm::toString(Xmm::CAL_THIN1);
	  else if(ranger(113., 118.))
	    new_value = Xmm::toString(Xmm::THIN1);
	  else if(ranger(162., 173.))
	    new_value = Xmm::toString(Xmm::CAL_THIN2);
	  else if(ranger(173., 178.))
	    new_value = Xmm::toString(Xmm::THIN2);
	  else if(ranger(222., 233.))
	    new_value = Xmm::toString(Xmm::CAL_MEDIUM);
	  else if(ranger(233., 238.))
	    new_value = Xmm::toString(Xmm::MEDIUM);
	  else if(ranger(282., 293.))
	    new_value = Xmm::toString(Xmm::CAL_THICK);
	  else if(ranger(293., 298.))
	    new_value = Xmm::toString(Xmm::THICK);
	  else if(ranger(-19.,-8.))
	    new_value = Xmm::toString(Xmm::CAL_OPEN);
	  else if(ranger(-8., -3.))
	    new_value = Xmm::toString(Xmm::OPEN);
	  else
	    throw ranger;
	  invalid = false;
	}
	catch(string &){
	  errstr << "I could not perform a lexical conversion on "
		 << p.name() << " with value " << p.value()
		 << ". The resulting housekeeping parameter " << nname
		 << " will not be created."
		 << Sas::warning("HousekeepingExtractionFailed");
	}
	catch(InRange &){
	  errstr << "The parameter " << p.name() << " with value "
		 << p.value() << " does not belong to any of the allowed ranges." << " The resulting housekeeping parameter " << nname << " will not be created."  << Sas::warning("HousekeepingExtractionFailed");
	}
	exp.insert_ippv(Ippv(nname, new_value, invalid));	
      }
    }    
  };


  // For PN exposures the MODE must be changed as follows:
  //  - if MODE == PrimeFullWindow && FRAME_TIME_PARAMETER != 0
  //      => MODE = PrimeFullWindowExtended
  //  - if MODE == PrimeFullWindowExtended && FRAME_TIME_PARAMETER == 0
  //      => MODE = PrimeFullWindow
  class EpnModeCorrector : public ActualTranslator {
  public:
    EpnModeCorrector() {}
    void operator()(OdfIngest::Exposure & exp) {
      using namespace OdfIngest;
      if(exp.has_ippv("MODE") && exp.has_ippv("FRAME_TIME_PARAMETER")){
	string new_mode;
	bool invalid = true;
	bool replace = false;
	Ippv p = exp.get_ippv("FRAME_TIME_PARAMETER");
	try {
	  string mode = exp.get_ippv("MODE").value();
	  int ftp = p.Ippv::value<int>();
	  if(mode == "PrimeFullWindow" && ftp != 0){
	    new_mode = "PrimeFullWindowExtended";
	    replace = true;
	  } else if(mode == "PrimeFullWindowExtended" && ftp == 0){
	    new_mode = "PrimeFullWindow";
	    replace = true;
	  }
	  invalid = false;
	} catch(string &) {
	  errstr << "I could not perform a lexical conversion on " << p.name()
		 << " with value " << p.value()
		 << ". I was trying to adjust the MODE for the exposure "
		 << exp.identifier()
		 << ". The correction fails and the MODE will not be set."
		 << Sas::warning("FIXME");
	  new_mode = Xmm::toString(Xmm::MODE_NONE);
	}
	if(replace){
	  exp.remove_ippv("MODE");
	  exp.insert_ippv(Ippv("MODE", new_mode, invalid));
	}
      }
    }
  };

  EmosGainTranslator emosGainTranslator;
  Emos1FilterTranslator emos1FilterTranslator;
  Emos2FilterTranslator emos2FilterTranslator;
  OmFilterTranslator omFilterTranslator;
  EpnFilterWheelTranslator epnFilterWheelTranslator;
  EpnFilterTranslator epnFilterTranslator;
  EpnModeCorrector epnModeCorrector;

};



template<class T> void OdfIngest::extract_hkparameters(Exposure & exp,
						       std::vector<Hkparameter *> & hkv, double hktimeoffset,
						       const T & hknamer)
{
  if(hkv.size() == 0)
    return;
  
  // sort by hk-file type to minimize the number of open/close
  std::sort(hkv.begin(), hkv.end(), SortByHkFileType());
  Xmm::DataMode dm = hkv[0]->hktype();
  
  using namespace SmartDal;
  DataSetH set = open_set(hknamer(exp, dm), Read, Low);
  TableH table = set->table(0); // HK tables don't have easy to guess
				// names, and in any case their name
				// encodes the extension number.
  for(unsigned k = 0; k < hkv.size(); ++k){
    Hkparameter * p = hkv[k];
    if(p->hktype() != dm){
      dm = p->hktype();
      set = SmartDal::open_set(hknamer(exp, dm), Read, Low);
      table = set->table(0);
    }
    try {
      SetExceptionErrorHandler x;
      errstr << "Extracting " << p->name() << " from column "
	     << p->colname() << ", periodic? " << p->periodic()
	     << Sas::message(Msg::AppLib, Msg::Noisy);
      p->process(table,
		 exp.start_time() + hktimeoffset,
		 exp.end_time() - hktimeoffset);
      Ippv ippv(p->name(), p->value());
      exp.insert_ippv(ippv);
      errstr << "Extracted value: " << p->value()
	     << Sas::message(Msg::AppLib, Msg::Noisy);
    }
    catch(SasException::Error e){
      errstr << "Exposure " << exp.identifier() << ": The extraction of the HK parameter "
	     << p->name() << "/" << p->colname()
	     << " failed with the following error:\n"
	     << e.code << ": " << e.message
	     << Sas::warning("HousekeepingExtractionFailed");
      Ippv ippv(p->name(), "" , true);
      exp.insert_ippv(ippv);      
    }
    catch(SasException::Fatal e){
      errstr << "Exposure " << exp.identifier() << ": The extraction of the HK parameter "
	     << p->name() << "/" << p->colname()
	     << " failed with the following error:\n"
	     << e.code << ": " << e.message
	     << Sas::warning("HousekeepingExtractionFailed");
      Ippv ippv(p->name(), "", true);
      exp.insert_ippv(ippv);      
    }
  }
}

template void OdfIngest::extract_hkparameters<HkNamer>(Exposure & exp,
						       std::vector<Hkparameter *> & hkv, double hktimeoffset,
						       const HkNamer & hknamer);

void OdfIngest::translate_hkparameters(Exposure & exp) {
  
  Xmm::Instrument ins = exp.instrument().identifier();
  
  switch(ins) {
  case Xmm::EMOS1:
    emosGainTranslator(exp);
    emos1FilterTranslator(exp);
    break;
  case Xmm::EMOS2:
    emosGainTranslator(exp);
    emos2FilterTranslator(exp);
    break;
  case Xmm::EPN:
    epnFilterWheelTranslator(exp);
    epnFilterTranslator(exp);
    epnModeCorrector(exp);
    break;
  case Xmm::RGS1:
    break;
  case Xmm::RGS2:
    break;
  case Xmm::OM:
    omFilterTranslator(exp);
    break;
  default:
    errstr << "In translate_hkparameters I cannot handle this instrument number: " << ins << Sas::error("UnknownInstrument");
  }
}
