// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine science & computing) 
// October 2001
//
// $Id: OdfIcd.cc,v 1.1 2003/05/06 08:53:04 gvacanti Exp $
//
#include <OdfIcd.h>

#include<string>
#include<map>

#include<sstream>
#include<errstr.h>

std::string OdfIcd::instrumentToString(const Xmm::Instrument& instrument)
{
    std::string name("XX");

  switch(instrument) {
  case Xmm::RGS1:
    {
      name = "R1";
      break;
    }
  case Xmm::RGS2:
    {
      name = "R2";
      break;
    }
  case Xmm::EPN:
    {
      name = "PN";
      break;
    }
  case Xmm::EMOS1:
    {
      name = "M1";
      break;
    }
  case Xmm::EMOS2:
    {
      name = "M2";
      break;
    }
  case Xmm::OM:
    {
      name = "OM";
      break;
    }
  case Xmm::ERM:
    {
      name = "RM";
      break;
    }
  default: {
      errstr << "The instrument with id " << instrument <<
      " is not recognized." << Sas::error("UnknownInstrument");
    }
  }
  return name;
}

std::string OdfIcd::modeToString(const Xmm::DataMode& datamode)
{

    std::string name("");
  switch(datamode) {
  case Xmm::PERIODIC_HOUSEKEEPING:
    {
      name = "PE";
      break;
    }
  case Xmm::HBR_CONFIG_HOUSEKEEPING:
    {
      name = "HC";
      break;
    }
  case Xmm::HBR_BUFFER_HOUSEKEEPING:
    {
      name = "HB";
      break;
    }
  case Xmm::HBR_THRESHOLD_HOUSEKEEPING:
    {
      name = "HT";
      break;
    }
  case Xmm::EXTRAHEATING_CONF_HOUSEKEEPING:
    {
      name = "EC";
      break;
    }
  case Xmm::THERMAL_HOUSEKEEPING:
    {
      name = "TM";
      break;
    }
  case Xmm::BRIGHT_PIXEL_HOUSEKEEPING:
    {
      name = "PT";
      break;
    }
  case Xmm::MAIN_PERIODIC_HOUSEKEEPING:
    {
      name = "PM";
      break;
    }
  case Xmm::ADD_PERIODIC_HOUSEKEEPING:
    {
      name = "PA";
      break;
    }
  case Xmm::FULL_PERIODIC_HOUSEKEEPING:
    {
      name = "PF";
      break;
    }
  case Xmm::TEMP_PERIODIC_HOUSEKEEPING:
    {
      name = "PC";
      break;
    }
  case Xmm::DPP_HOUSEKEEPING:
    {
      name = "DP";
      break;
    }
  case Xmm::NON_PERIODIC_HOUSEKEEPING:
    {
      name = "NP";
      break;
    }
  case Xmm::DPP_HOUSEKEEPING_1:
    {
      name = "D1";
      break;
    }
  case Xmm::DPP_HOUSEKEEPING_2:
    {
      name = "D2";
      break;
    }
  case Xmm::HK_P1:
    {
      name = "P1";
      break;
    }
  case Xmm::HK_P2:
    {
      name = "P2";
      break;
    }
  case Xmm::HK_P3:
    {
      name = "P3";
      break;
    }
  case Xmm::HK_P4:
    {
      name = "P4";
      break;
    }
  case Xmm::HK_P5:
    {
      name = "P5";
      break;
    }
  case Xmm::HK_P6:
    {
      name = "P6";
      break;
    }
  case Xmm::HK_P7:
    {
      name = "P7";
      break;
    }
  case Xmm::HK_P8:
    {
      name = "P8";
      break;
    }
  case Xmm::HK_P9:
    {
      name = "P9";
      break;
    }
  default:
    {      
      errstr << "The data mode " << datamode <<
	" is not recognized here." << Sas::error("UnknownDataMode");
    }
  }
  return name;
}

std::string OdfIcd::hkname(unsigned revolution, const std::string & observation,
		   Xmm::Instrument i, Xmm::DataMode d)
{
    stringstream s;
    s << std::setw(4) << std::setfill('0') << revolution << "_"
      << observation << "_" << OdfIcd::instrumentToString(i) +
	"X00000" + OdfIcd::modeToString(d) + "H.FIT";
  return s.str();
}

std::string OdfIcd::hkname(unsigned revolution, unsigned observation,
		   Xmm::Instrument i, Xmm::DataMode d)
{
    stringstream s;
    s << std::setw(10) << std::setfill('0') << observation;
    return hkname(revolution, s.str(), i, d);
}

std::string OdfIcd::hkname(const OdfIngest::Exposure & e, Xmm::DataMode d) {
  const OdfIngest::Observation * obs = e.observation();
  return hkname(obs->revolution(), obs->identifier(),
		e.instrument().identifier(), d);
}
