// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// June 20, 2000
//
#include "IppvImplementation.h"
#include "IppvTranslator.h"

#include <Tracer.h>
#include <errstr.h>
#include <SasDalUtils.h>
#include <DataFilter.h>
#include <cstdio>
#include <dal_utilities.h>

#include <string>
#include <map>
#include <typeinfo>

using namespace std;
//
// IppvServer
//

IppvServer::IppvServer() { PROC ENDPROC }

IppvServer::~IppvServer() { PROC ENDPROC }

Ippv * IppvServer::ippv(const string& name,
			const string& colname,
			const Dal::DataType datatype,
			const string& label,
			const Ippv::Type type,
			const bool& singlevalue) const
{
  PROC
  Ippv * r = 0;
  switch(type) {
  case Ippv::Periodic:
    switch(datatype) {
    case Dal::Int8:
      r = new PeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int16:
      r = new PeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint16:
      r = new PeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int32:
      r = new PeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint32:
      r = new PeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real32:
      r = new PeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real64:
      r = new PeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, singlevalue);
      break;
    case Dal::DString:
      r = new PeriodicIppv<string, char, &Data::stringData>(name, colname, label, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;

  case Ippv::NonPeriodic:
    switch(datatype) {
    case Dal::Int8:
      r = new NonPeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int16:
      r = new NonPeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint16:
      r = new NonPeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, singlevalue);
      break;
    case Dal::Int32:
      r = new NonPeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Uint32:
      r = new NonPeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real32:
      r = new NonPeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, singlevalue);
      break;
    case Dal::Real64:
      r = new NonPeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, singlevalue);
      break;
    case Dal::DString:
      r = new NonPeriodicIppv<string, char, &Data::stringData>(name, colname, label, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  default:
    Sas::error("UnsupportedHousekeepingParameterType");
    break;
  }
  RETURN r;
}

Ippv * IppvServer::ippv(const string& name,
			const string& colname,
			const Dal::DataType datatype,
			const string& label,
			const Ippv::Type type,
			const string& expression,
			const bool& singlevalue) const
{
  PROC
  Ippv * r = 0;
  switch(type) {
  case Ippv::Periodic:
    switch(datatype) {
    case Dal::Int8:
      r = new PeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int16:
      r = new PeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint16:
      r = new PeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int32:
      r = new PeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint32:
      r = new PeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real32:
      r = new PeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real64:
      r = new PeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::DString:
      r = new PeriodicIppv<string, char , &Data::stringData>(name, colname, label, expression, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  case Ippv::NonPeriodic:
    switch(datatype) {
    case Dal::Int8:
      r = new NonPeriodicIppv<int8, int8, &Data::int8Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int16:
      r = new NonPeriodicIppv<int16, int16, &Data::int16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint16:
      r = new NonPeriodicIppv<uint16, uint16, &Data::uint16Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Int32:
      r = new NonPeriodicIppv<int32, int32, &Data::int32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Uint32:
      r = new NonPeriodicIppv<uint32, uint32, &Data::uint32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real32:
      r = new NonPeriodicIppv<real32, real32, &Data::real32Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::Real64:
      r = new NonPeriodicIppv<real64, real64, &Data::real64Data>(name, colname, label, expression, singlevalue);
      break;
    case Dal::DString:
      r = new NonPeriodicIppv<string, char , &Data::stringData>(name, colname, label, expression, singlevalue);
      break;
    default:
      errstr << "Cannot handle type " << static_cast<int>(datatype)
	     << " here" << Sas::error("UnknownDalDataType");
    }
    break;
  default:
    Sas::error("UnsupportedHousekeepingParameterType");
    break;
  }
  RETURN r;
}

//
// IppvCommon
//
template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::IppvCommon(const string& name,
				   const string& colname,
				   const string& label,
				   const bool& singlevalue)
  : _name(name),
    _colname(colname),
    _expression(""),
    _label(label),
    _filter(false),
    _singlevalue(singlevalue),
    _has_translator(false),
    _trans(new IppvTranslator<T>)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::IppvCommon(const string& name,
				   const string& colname,
				   const string& label,
				   const string& expression,
				   const bool& singlevalue)
  : _name(name),
    _colname(colname),
    _expression(expression),
    _label(label),
    _filter(true),
    _singlevalue(singlevalue),
    _has_translator(false),
    _trans(new IppvTranslator<T>)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
IppvCommon<T, S, data>::~IppvCommon() { PROC ENDPROC }

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::fetch(const TimeOrderedTable *tab,
				   const STime& start,
				   const STime& stop) throw(IppvException)
{
  PROC

  unsigned rowl = lower_row(tab, start);
  if(rowl == tab->rows()){
    errstr << tab->table().qualifiedName() + " Time range: "
	   << tab->time_of_row(0) << " - "
	   << tab->time_of_row(tab->rows() - 1)
	   << endl << "Time sought: " << start
	   << Sas::message(Msg::AppLib, Msg::Noisy);
    stringstream s;
    s << "TimeOutOfRange: lower time bound for " + _name + " in " +
      tab->table().qualifiedName() +
      " not available. Looking for an entry at or around "
      << start << " but the available time range in "
      <<  tab->table().name() << " is "
      << tab->time_of_row(0) << " - "
      << tab->time_of_row(tab->rows() - 1);
    throw OutOfRange(s.str());
  }
  unsigned rowu = upper_row(tab, stop);
  if(rowu == tab->rows()){
    stringstream s;
    s << "TimeOutOfRange: upper time bound for " + _name + " in " +
      tab->table().qualifiedName() +
      " not available. Looking for an entry at or around "
      << stop << " but the available time range in "
      <<  tab->table().name() << " is "
      << tab->time_of_row(0) << " - "
      << tab->time_of_row(tab->rows() - 1);
    throw OutOfRange(s.str());
  }

  vector<DataFilter::RowRange> rr;
  if(_filter){
    stringstream s;
    // Perverse!! #ROW start from 1 but the DAL starts from 0
    s << "(" + _expression + ") && (#ROW >= " << rowl+1
      << " && #ROW <= " << rowu+1 << ")";
    string expression = s.str();
    DataFilter df(expression);
    df.doSelect(tab->table(), rr);
    if(rr.size() == 0)
      {
	throw NoHkRecords("NoAvailableRows: No rows match the expression "
			  + expression + " in " + tab->table().qualifiedName());
      }
  } else {
    DataFilter::RowRange r;
    r.startRow = rowl;
    r.stopRow = rowu;
    rr.push_back(r);
  }
  
  
  errstr << "start/stop " << start << " / " << stop  << Sas::message(Msg::AppLib, Msg::Noisy);
  for(unsigned k=0; k<rr.size(); ++k)
    {
      errstr << "Row range = [" << k << "] " << rr[k].startRow << " - " << rr[k].stopRow
	     << Sas::message(Msg::AppLib, Msg::Noisy);
    }
  
  
  //  _label += " [" + tab->table().column(_colname)->units() + "]";
  try {
    //      if(_singlevalue && rr.size() > 1){
    // setValue below does not know anything about _singlevalue. So we
    // make sure that the row range is mododified here.
    if(_singlevalue){
      rr.erase(rr.begin() + 1, rr.end());
      rr[0].stopRow = rr[0].startRow;
      Sas::message(Msg::AppLib, Msg::Noisy, "RowRange modified.");
    }
    setValue(tab->table().column(_colname), rr);
  }
  catch (IppvException &e) {
    throw;
  }

  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
unsigned IppvCommon<T, S, data>::lower_row(const TimeOrderedTable *,
					   const STime&) const
{
  PROC
  Sas::error("NotImplementedHere", "this call is not implemented here");
  RETURN 0;
}

template<class T, class S, S * (Data::*data)() const>
unsigned IppvCommon<T, S, data>::upper_row(const TimeOrderedTable *,
				       const STime&) const
{
  PROC
  Sas::error("NotImplementedHere", "this call is not implemented here");
  RETURN 0;
}



template<class T> static string dalTypeToString(T v)
{
  static bool firstTime = true;
  static map<string, string> m;
  if(firstTime){
    m[typeid(bool).name()] = toString(Dal::Bool);
    m[typeid(int8).name()] = toString(Dal::Int8);
    m[typeid(uint16).name()] = toString(Dal::Uint16);
    m[typeid(int16).name()] = toString(Dal::Int16);
    m[typeid(uint32).name()] = toString(Dal::Uint32);
    m[typeid(int32).name()] = toString(Dal::Int32);
    m[typeid(real32).name()] = toString(Dal::Real32);
    m[typeid(real64).name()] = toString(Dal::Real64);
    m[typeid(dstring).name()] = toString(Dal::DString);
    m[typeid(string).name()] = "string";
    firstTime = false;
  }
  if(m.find(typeid(v).name()) != m.end())
    return m[typeid(v).name()];
  else
    return typeid(v).name();
}
// FIXME: duplicated code in the specialization

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::print(ostream& os, const bool verbose) const
{
  PROC
  if(verbose){
    os << _name << " " << dalTypeToString<T>(_value) << " [" << _colname ;
    if(_filter)
      os << " subject to " << _expression;
    os << "]";
    if(_has_translator){
      os << endl << "This parameter is translated as follows:" << endl;
      _trans->print(os);
    }
  } else {
    os << _name << " = ";
    if(_has_translator){
      try{
    	  os << _trans->look_up(_value);
      }
      catch(NoTranslationAvailable &e){
    	  e.print(errstr);
    	  errstr << "It is not possible to translate " + _name
    			  + ". Its raw value is " << _value
				  << Sas::warning("TranslationNotAvailable");
    	  os << "UNDEFINED // Raw = " << _value;
      }
    }
    else
      os << _value;
    os << " // "  << _label;
  }
  os << endl;
  ENDPROC
}
//AI: Filter identification
// Writing of the FILTER value
template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::print2(ostream& os, string k, float v)
{
  PROC
  stringstream ss;
  errstr << "Something happened with the filter identification using the potentiometer in F1122. "
	 << "The value is [";
  if(_has_translator){
    try{
  	  ss << _trans->look_up(_value);
    }
    catch(NoTranslationAvailable &e){
  	  e.print(errstr);
  	  errstr << "It is not possible to translate " + _name
  			  + ". Its raw value is " << _value
				  << Sas::warning("TranslationNotAvailable");
  	  ss << "UNDEFINED // Raw = " << _value;
    }
  }
  _value = v;
  errstr  << ss.str() << "] but according to FD130 should be [" << _trans->look_up(_value) << "]" << Sas::message(Msg::App, Msg::Sparse);
  stringstream sss;
  sss << "Filter position translated from FD130 column ";

  _label = sss.str();
  //cout << "NEWFILTER " <<_trans->look_up(_value) << endl;
  ENDPROC
}

//AI: Filter Identification
template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::print3(ostream& os, string k, float v)
{
  PROC
    //stringstream ss;
    //ss << "Filter potentiometer value "<< _name  << " " << _value << " " << _label;

    //ss  << "  name  " << k << " valuee " << std::setprecision(3)<< v << endl;
    //cout << ss.str() << endl;
  stringstream sss;
  sss << "Previous potentiometer value was " << _value << " " << _label;

  _label = sss.str();
  _value = v;
  //cout << "NEWFILTER WHEEL POSITION " <<_value << endl;
  ENDPROC
}

// Specialization required because in the DAL int8 = char and << will
// try to print a character instead of an unsigned integer
template<>
void IppvCommon<int8, int8, &Data::int8Data>::print(ostream& os, const bool verbose) const
{
  PROC
  if(verbose){
    os << _name << " " << dalTypeToString<int8>(_value) << " [" << _colname ;
    if(_filter)
      os << " subject to " << _expression;
    os << "]";
    if(_has_translator){
      os << "This parameter is translated as follows:" << endl;
      _trans->print(os);
    }
  } else {
    os << _name << " = ";
    if(_has_translator){
      try{
	os << _trans->look_up(_value);
      }
      catch(NoTranslationAvailable &e){
	e.print(errstr);
	errstr << "It is not possible to translate " + _name
	  + ". Its raw value is " << _value
	       << Sas::warning("TranslationNotAvailable");
	throw;
      }
    }
    else
      os << static_cast<int>(_value);
    os << " // " << _label;
  }
  os << endl;
  ENDPROC
}


// SCR117
template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::value(string &val) const
{
  PROC
    
    stringstream s;
  if(_has_translator)
    {
      try
	{
	  s << _trans->look_up(_value);
	}
      catch(NoTranslationAvailable e)
	{
	  e.print(errstr);
	  errstr << "It is not possible to translate " + _name
	    + ". Its raw value is " << _value
		 << Sas::warning("TranslationNotAvailable");
	  s << "SAS_UNDEFINED"; 
	}
    }
  else 
    {
      s << _value;
    }
  
  val = s.str();
  
  ENDPROC
	
}

// Specialization required because in the DAL int8 = char and << will
// try to print a character instead of an unsigned integer
template<>
void IppvCommon<int8, int8, &Data::int8Data>::value(string &val) const
{

	PROC

	stringstream s;

	if(_has_translator)
	{
		try
		{
			s << _trans->look_up(_value);
		}
		catch(NoTranslationAvailable &e)
		{
			e.print(errstr);
			errstr << "It is not possible to translate " + _name
				+ ". Its raw value is " << _value
				<< Sas::warning("TranslationNotAvailable");
			s << "SAS_UNDEFINED"; 
		}
	}
    else 
	{
      s << static_cast<int>(_value);
	}

	val = s.str();

	ENDPROC
}
// END SCR117


template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::notAvailable(ostream& os) const
{
  PROC
  os << "// WARNING: " << _name << " not found" << endl;
  os << _name + "= Unknown // " + _label << endl;
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
const string& IppvCommon<T, S, data>::name() const
{
  PROC
    RETURN _name;
}

template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T, S, data>::setValue(const Column *column,
				      const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  // ATTENTION: column->data() returns ColumnData*, but the pointer to
  // member stuff only works for the base class Data. I do not know if
  // this is a bug in gcc of it is supposed to be this way.
  Data* cd = column->data();
  _value = (cd->*data)()[r[0].startRow];
  unsigned nmismatch = 0;
  for(unsigned i = 0; i < r.size(); ++i)
    {
      //AI:      cout <<"column "<< _colname << " parameter name "<< _name <<  " start row " <<  r[i].startRow << " stop row " << r[i].stopRow << endl;
      for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
	T v;
	if((v = (cd->*data)()[k]) != _value)
    	  ++nmismatch;
	// tolerate one HK record with a different: value      
	if(nmismatch > 1){
    	  stringstream s;
    	  s << "Parameter " + _name + " from column " + _colname
	    << ". Expected " << _value << " but found " << v
	    << " at row " << k << endl;
    	  delete cd;
    	  throw ParameterChange(s.str());
	}
      }
    }
  delete cd;
  ENDPROC
}

// SCR117
// Specialization required because in the DAL int8 = char and << will
// try to print a character instead of an unsigned integer
template<>
void IppvCommon<int8, int8, &Data::int8Data>::setValue(const Column *column,
				      const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  // ATTENTION: column->data() returns ColumnData*, but the pointer to
  // member stuff only works for the base class Data. I do not know if
  // this is a bug in gcc of it is supposed to be this way.
  Data* cd = column->data();
  _value = (cd->*&Data::int8Data)()[r[0].startRow];
  unsigned nmismatch = 0;

  for(unsigned i = 0; i < r.size(); ++i)
    for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
      int8 v;
      if((v = (cd->*&Data::int8Data)()[k]) != _value)
    	  ++nmismatch;
      // tolerate one HK record with a different value      
      if(nmismatch > 1){
	stringstream s;
	s << "Parameter " + _name + " from column " + _colname
	  << ". Expected " << static_cast<int>(_value) << " but found " << static_cast<int>(v)
	  << " at row " << k;
	delete cd;
	throw ParameterChange(s.str());
      }
    }
  delete cd;
  ENDPROC
}


// Specialization for real-valued cases, when we calculate the mean of
// the values in the range. 

template<>
void IppvCommon<real32, real32, &Data::real32Data>::setValue(const Column *column,
							     const vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  // ATTENTION: column->data() returns ColumnData*, but the pointer to
  // member stuff only works for the base class Data. I do not know if
  // this is a bug in g++ of it is supposed to be this way.
  Data* cd = column->data();
  real32 v = 0.;
  unsigned nrows = 0;
  for(unsigned i = 0; i < r.size(); ++i){
    for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
      ++nrows;
      v += (cd->*&Data::real32Data)()[k];
    }
  }
  _value = v/nrows;
  delete cd;
  ENDPROC
}

// Specialization required because reading rows from a character
// column requires a different mechanism.
template<>
void IppvCommon<std::string, char, &Data::stringData>::setValue(const Column *column,
							   const std::vector<DataFilter::RowRange>& r)
  throw(IppvException) {
  PROC
  // ATTENTION: It assumes that there is at least one component in r.
  assert(r.size() > 0);
  ColumnData* cd = column->data();
  _value = SasDalUtils::getStringColumnRow(cd, r[0].startRow);
  unsigned nmismatch = 0;
  int rows = 0;

  for(unsigned i = 0; i < r.size(); ++i)
    {
      //AI: Filter identification
      if (r[i].stopRow >= r[i].startRow)
	{
	  if ((_name == "FILTER_STRING") || (_name == "FILTER_FD130_STRING") ) {
	    
	    rows = int(r[i].stopRow) - int(r[i].startRow) ;
	    //cout << "_name " << _name << " row : " << rows << " start "<< r[i].startRow << " stop "<<  r[i].stopRow << endl;
	  }
	  for(unsigned k = r[i].startRow; k <= r[i].stopRow; ++k){
	    string v;
	    if((v = SasDalUtils::getStringColumnRow(cd, k)) != _value)
	      {
		++nmismatch;
	      }
	    // tolerate one HK record with a different value
	    if ((_name != "FILTER_STRING") && (_name != "FILTER_FD130_STRING") )
	      {
		if(nmismatch > 1)
		  {
		    stringstream s;
		    s << "Parameter " + _name + " from column " + _colname
		      << ". Expected " << _value << " but found " << v
		      << " at row " << k;
		    delete cd;
		    throw ParameterChange(s.str());
		  }
	      }
	  }	
	  //Filter Identification
	  if ((_name == "FILTER_STRING") || (_name == "FILTER_FD130_STRING") )
	    {
	      if (rows != 0)
		{
		  double perc =  (double(rows) - double(nmismatch))/double(rows);
		  //cout << "Column "<< _name  << " nmismatch "  << nmismatch << " rows "<< rows << " perc "<< perc << endl;
		  if (perc < 0.85 )
		    {
		      errstr << "Too many entries with diferent values for column  " << _name << Sas::warning("TooManyDifferences");
		      _value = "FILTER_UNDEF";
		    }
		}
	    }
	}
    }

  delete cd;
  ENDPROC
}




template<class T, class S, S * (Data::*data)() const>
void IppvCommon<T,S,data>::addTranslator(IppvTranslator<T> *t)
{
  PROC
  _trans = t;
  _has_translator = true;
  ENDPROC
}


//
// PeriodicIppv
//
template<class T, class S, S * (Data::*data)() const>
PeriodicIppv<T, S, data>::PeriodicIppv(const string& name,
				       const string& colname,
				       const string& label,
				       const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, singlevalue)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
PeriodicIppv<T, S, data>::PeriodicIppv(const string& name,
				       const string& colname,
				       const string& label,
				       const string& expression,
				       const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, expression, singlevalue)
{
  PROC
  ENDPROC
}

//  template<class T, class S, S * (Data::*data)() const>
//  PeriodicIppv<T, S, data>::~PeriodicIppv() { PROC ENDPROC }
 
template<class T, class S, S * (Data::*data)() const>
unsigned PeriodicIppv<T, S, data>::lower_row(const TimeOrderedTable *tab,
					     const STime& t) const
{
  PROC
  RETURN tab->lower_bound(t);

}

template<class T, class S, S * (Data::*data)() const>
unsigned PeriodicIppv<T, S, data>::upper_row(const TimeOrderedTable *tab,
					     const STime& t) const
{
  PROC
  RETURN tab->upper_bound(t);

}

//
// NonPeriodicIppv
//
template<class T, class S, S * (Data::*data)() const>
NonPeriodicIppv<T, S, data>::NonPeriodicIppv(const string& name,
					     const string& colname,
					     const string& label,
					     const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, singlevalue)
{
  PROC
  ENDPROC
}

template<class T, class S, S * (Data::*data)() const>
NonPeriodicIppv<T, S, data>::NonPeriodicIppv(const string& name,
					     const string& colname,
					     const string& label,
					     const string& expression,
					     const bool& singlevalue)
  : IppvCommon<T,S,data>(name, colname, label, expression, singlevalue)
{
  PROC
  ENDPROC
}

//  template<class T, class S, S * (Data::*data)() const>
//  NonPeriodicIppv<T, S, data>::~NonPeriodicIppv() { PROC ENDPROC }
 
template<class T, class S, S * (Data::*data)() const>
unsigned NonPeriodicIppv<T, S, data>::lower_row(const TimeOrderedTable *tab,
						const STime& t) const
{
  PROC
  RETURN tab->first_less_than(t);
}

template<class T, class S, S * (Data::*data)() const>
unsigned NonPeriodicIppv<T, S, data>::upper_row(const TimeOrderedTable *tab,
						const STime& t) const
{
  PROC
  RETURN tab->last_less_than(t);
}

//

template void IppvCommon<uint16, uint16, &Data::uint16Data>::addTranslator(IppvTranslator<uint16> *);
template void IppvCommon<int16, int16, &Data::int16Data>::addTranslator(IppvTranslator<int16> *);
template void IppvCommon<std::string, char, &Data::stringData>::addTranslator(IppvTranslator<std::string> *);
template void IppvCommon<real32, real32, &Data::real32Data>::addTranslator(IppvTranslator<real32> *);


// These are needed on DEC/OSF
template class PeriodicIppv<uint16, uint16, &Data::uint16Data>;
template class PeriodicIppv<std::string, char,   &Data::stringData>;
template class PeriodicIppv<int16,  int16,  &Data::int16Data>;
template class PeriodicIppv<real32, real32, &Data::real32Data>;
